#include <assert.h>
#include <gd.h>
#include <stdlib.h>
#include "Grafx.h"


/* -----------------------------------------------------------------------
 *    Private Functions
 * -----------------------------------------------------------------------
 */
void _setColors(Grafx *g) {
    
  g->colors.bgcolor = gdImageColorAllocate(g->im, 0,   0, 0);

  /* coloring function: morph
   *    - morphology corresponds to r, y, b
   */

  g->colors.stolon  = gdImageColorAllocate(g->im, 255, 63, 63);
  g->colors.fused   = gdImageColorAllocate(g->im, 255, 255, 63);
  g->colors.unfused = gdImageColorAllocate(g->im, 63, 63, 255);

  
  /* coloring function: morphgeno
   *    - genotype   corresponds to different luminosity
   *    - morphology corresponds to different color class
   */

  g->colors.stolon_xx_xx = gdImageColorAllocate(g->im, 255, 0  , 0  );
  g->colors.stolon_xx_xy = gdImageColorAllocate(g->im, 255, 22 , 22 );
  g->colors.stolon_xx_yy = gdImageColorAllocate(g->im, 255, 44 , 44 );
  g->colors.stolon_xy_xx = gdImageColorAllocate(g->im, 255, 66 , 66 );
  g->colors.stolon_xy_xy = gdImageColorAllocate(g->im, 255, 88 , 88 );
  g->colors.stolon_xy_yy = gdImageColorAllocate(g->im, 255, 110, 110);
  g->colors.stolon_yy_xx = gdImageColorAllocate(g->im, 255, 132, 132);
  g->colors.stolon_yy_xy = gdImageColorAllocate(g->im, 255, 154, 154);
  g->colors.stolon_yy_yy = gdImageColorAllocate(g->im, 255, 176, 176);
   
  g->colors.fused_xx_xx = gdImageColorAllocate(g->im, 224, 224, 0  );
  g->colors.fused_xx_xy = gdImageColorAllocate(g->im, 228, 228, 20 );
  g->colors.fused_xx_yy = gdImageColorAllocate(g->im, 232, 232, 40 );
  g->colors.fused_xy_xx = gdImageColorAllocate(g->im, 236, 236, 60 );
  g->colors.fused_xy_xy = gdImageColorAllocate(g->im, 240, 240, 80 );
  g->colors.fused_xy_yy = gdImageColorAllocate(g->im, 244, 244, 100);
  g->colors.fused_yy_xx = gdImageColorAllocate(g->im, 248, 248, 120);
  g->colors.fused_yy_xy = gdImageColorAllocate(g->im, 252, 252, 140);
  g->colors.fused_yy_yy = gdImageColorAllocate(g->im, 255, 255, 160);
  
  g->colors.unfused_xx_xx = gdImageColorAllocate(g->im, 0  , 22 , 255);
  g->colors.unfused_xx_xy = gdImageColorAllocate(g->im, 24 , 44 , 255);
  g->colors.unfused_xx_yy = gdImageColorAllocate(g->im, 48 , 66 , 255);
  g->colors.unfused_xy_xx = gdImageColorAllocate(g->im, 72 , 88 , 255);
  g->colors.unfused_xy_xy = gdImageColorAllocate(g->im, 96 , 110, 255);
  g->colors.unfused_xy_yy = gdImageColorAllocate(g->im, 120, 132, 255);
  g->colors.unfused_yy_xx = gdImageColorAllocate(g->im, 144, 154, 255);
  g->colors.unfused_yy_xy = gdImageColorAllocate(g->im, 168, 176, 255);
  g->colors.unfused_yy_yy = gdImageColorAllocate(g->im, 192, 198, 255);

  
  /* coloring function: genomorph
   *    - genotype   corresponds to different color class
   *    - morphology corresponds to different luminosity 
   */
  
}


/* -----------------------------------------------------------------------
 *    Public Functions
 * -----------------------------------------------------------------------
 */

Grafx *newGrafx(Opts *opts) {
  int   length = opts->size * opts->pixelscale;
  Grafx *g     = malloc(sizeof(Grafx));
  g->coloring = opts->coloring;
  g->im = gdImageCreate(length, length);
  _setColors(g);
  return g;  
}


void world2png(Opts *opts, World w, Grafx *g) {
  int x, y, color = 0;
  for (x = 0; x < opts->size; x++)
    for (y = 0; y < opts->size; y++) {
      
      if (w[x][y].alive)

        if (g->coloring == morph) {
          switch (w[x][y].morph) {
            case stolon:
              color = g->colors.stolon;
              break;
            case unfused:
              color = g->colors.unfused;
              break;
            case fused:
              color = g->colors.fused;
              break;
            default:
              /* should not get here */
              assert(1);
          }
        }
        else {

          if (g->coloring == morphgeno)
            switch (w[x][y].morph) {

              case stolon:
                switch (w[x][y].genome[0]) {
                  case xx:
                    switch (w[x][y].genome[1]) {
                      case xx:
                        color = g->colors.stolon_xx_xx;
                        break;
                      case xy:
                        color = g->colors.stolon_xx_xy;
                        break;
                      case yy:
                        color = g->colors.stolon_xx_yy;
                        break;
                    }
                    break;
                  case xy:
                    switch (w[x][y].genome[1]) {
                      case xx:
                        color = g->colors.stolon_xy_xx;
                        break;
                      case xy:
                        color = g->colors.stolon_xy_xy;
                        break;
                      case yy:
                        color = g->colors.stolon_xy_yy;
                        break;
                    }
                    break;
                  case yy:
                    switch (w[x][y].genome[1]) {
                      case xx:
                        color = g->colors.stolon_yy_xx;
                        break;
                      case xy:
                        color = g->colors.stolon_yy_xy;
                        break;
                      case yy:
                        color = g->colors.stolon_yy_yy;
                        break;
                    }
                    break;
                }
                break;

              case fused:
                switch (w[x][y].genome[0]) {
                  case xx:
                    switch (w[x][y].genome[1]) {
                      case xx:
                        color = g->colors.fused_xx_xx;
                        break;
                      case xy:
                        color = g->colors.fused_xx_xy;
                        break;
                      case yy:
                        color = g->colors.fused_xx_yy;
                        break;
                    }
                    break;
                  case xy:
                    switch (w[x][y].genome[1]) {
                      case xx:
                        color = g->colors.fused_xy_xx;
                        break;
                      case xy:
                        color = g->colors.fused_xy_xy;
                        break;
                      case yy:
                        color = g->colors.fused_xy_yy;
                        break;
                    }
                    break;
                  case yy:
                    switch (w[x][y].genome[1]) {
                      case xx:
                        color = g->colors.fused_yy_xx;
                        break;
                      case xy:
                        color = g->colors.fused_yy_xy;
                        break;
                      case yy:
                        color = g->colors.fused_yy_yy;
                        break;
                    }
                    break;
                }
                break;

              case unfused:
                switch (w[x][y].genome[0]) {
                  case xx:
                    switch (w[x][y].genome[1]) {
                      case xx:
                        color = g->colors.unfused_xx_xx;
                        break;
                      case xy:
                        color = g->colors.unfused_xx_xy;
                        break;
                      case yy:
                        color = g->colors.unfused_xx_yy;
                        break;
                    }
                    break;
                  case xy:
                    switch (w[x][y].genome[1]) {
                      case xx:
                        color = g->colors.unfused_xy_xx;
                        break;
                      case xy:
                        color = g->colors.unfused_xy_xy;
                        break;
                      case yy:
                        color = g->colors.unfused_xy_yy;
                        break;
                    }
                    break;
                  case yy:
                    switch (w[x][y].genome[1]) {
                      case xx:
                        color = g->colors.unfused_yy_xx;
                        break;
                      case xy:
                        color = g->colors.unfused_yy_xy;
                        break;
                      case yy:
                        color = g->colors.unfused_yy_yy;
                        break;
                    }
                    break;
                }
                break;
            }
        }

      else
        color = g->colors.bgcolor;

      if (opts->pixelscale == 2) {
        /* 4 block "pixels" */
        gdImageSetPixel(g->im, 2 * x,     2 * y,     color);
        gdImageSetPixel(g->im, 2 * x + 1, 2 * y,     color);
        gdImageSetPixel(g->im, 2 * x,     2 * y + 1, color);
        gdImageSetPixel(g->im, 2 * x + 1, 2 * y + 1, color);
      }
      else
        gdImageSetPixel(g->im, x, y, color);
      
    }
      
}


void writepng(Grafx *g, char* fname) {
  FILE *pngout = fopen(fname, "wb");
  gdImagePng(g->im, pngout);
  fclose(pngout);
}

