#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include "World.h"
#include "rngmit.h"


/* -----------------------------------------------------------------------
 *    Private Functions
 * ----------------------------------------------------------------------- */


/* utility function to make exact copy of hydractinia */
void _copyHydra(Hydractina *dest, Hydractina *src) {
  dest->alive     = src->alive;
  dest->round     = src->round;
  dest->where.x   = src->where.x;
  dest->where.y   = src->where.y;
  dest->genome[0] = src->genome[0];
  dest->genome[1] = src->genome[1];
  dest->morph     = src->morph;
}


/* returns the number of friendly feeding cells */
int _countMatCells(Opts *opts, World w, Hydractina *hydra) {
  int x, y, matCellCount = 0;
  /* (0,0) is the top-left corner */

  /* count the number of mat cells nearby (in a stupid way) */
  x = hydra->where.x;    y = hydra->where.y - 1;
  if (y>=0 && _alloRecognize(hydra, w[x][y]) && _isMat(w[x][y]))
      matCellCount++;

  x = hydra->where.x;    y = hydra->where.y - 2;
  if (y>=0 && _alloRecognize(hydra, w[x][y]) && _isMat(w[x][y]))
      matCellCount++;

  x = hydra->where.x-1;  y = hydra->where.y - 1;
  if (x>=0 && y>=0 && _alloRecognize(hydra, w[x][y]) && _isMat(w[x][y]))
      matCellCount++;

  x = hydra->where.x-1;  y = hydra->where.y;
  if (x>=0 && _alloRecognize(hydra, w[x][y]) && _isMat(w[x][y]))
      matCellCount++;

  x = hydra->where.x-2;  y = hydra->where.y;
  if (x>=0 && _alloRecognize(hydra, w[x][y]) && _isMat(w[x][y]))
      matCellCount++;

  x = hydra->where.x-1;  y = hydra->where.y + 1;
  if (x>=0 && y<opts->size && _alloRecognize(hydra, w[x][y]) && _isMat(w[x][y]))
      matCellCount++;

  x = hydra->where.x;    y = hydra->where.y + 1;
  if (y<opts->size && _alloRecognize(hydra, w[x][y]) && _isMat(w[x][y]))
      matCellCount++;

  x = hydra->where.x;    y = hydra->where.y + 2;
  if (y<opts->size && _alloRecognize(hydra, w[x][y]) && _isMat(w[x][y]))
      matCellCount++;

  x = hydra->where.x+1;  y = hydra->where.y + 1;
  if (x<opts->size && y<opts->size && _alloRecognize(hydra, w[x][y]) && _isMat(w[x][y]))
      matCellCount++;

  x = hydra->where.x+1;  y = hydra->where.y;
  if (x<opts->size && _alloRecognize(hydra, w[x][y]) && _isMat(w[x][y]))
      matCellCount++;

  x = hydra->where.x+2;  y = hydra->where.y;
  if (x<opts->size && _alloRecognize(hydra, w[x][y]) && _isMat(w[x][y]))
      matCellCount++;

  x = hydra->where.x+1;  y = hydra->where.y-1;
  if (x<opts->size && y>=0 && _alloRecognize(hydra, w[x][y]) && _isMat(w[x][y]))
      matCellCount++;
  /* done being stupid */

  return matCellCount;
}


/* where (if anywhere) to reproduce */
int _fuck(Opts *opts, World w, Hydractina* hydra) {
  int i, x = -1, y = -1;
  int rn = floor(rngmit * 4);
  int matCellCount = _countMatCells(opts, w, hydra);
  
  /* check if fucking self in a productive way */
  if (rngmit < (opts->p_fuck_base + matCellCount*opts->p_fuck_mat)) {
    rn = floor(rngmit * 4);
    
    for (i = 0 ; i < 4; i ++, rn = (rn + i) % 4) {
      
      switch (rn) {
        case 0:
          /* above */
          x = hydra->where.x;
          y = hydra->where.y - 1;
          if (y < 0)
            continue;
          break;
        case 1:
          /* right */
          x = hydra->where.x + 1;
          y = hydra->where.y;
          if (x >= (opts->size))
            continue;
          break;
        case 2:
          /* below */
          x = hydra->where.x;
          y = hydra->where.y + 1;
          if (y >= (opts->size))
            continue;
          break;
        case 3:
          /* left */
          x = hydra->where.x - 1;
          y = hydra->where.y;
          if (x < 0)
            continue;
          break;
      }
      
      /* if dead neighbor, copy self into space */
      if( !w[x][y].alive ) {
        _copyHydra(&w[x][y], hydra);
        w[x][y].where.x = x;
        w[x][y].where.y = y;
        
        /* morphology mutation */
				if (rngmit < opts->p_mut_morph)
					w[x][y].morph = floor(rngmit*3);
				if (rngmit < opts->p_mut_morph)
					hydra->morph  = floor(rngmit*3);

        /* genome mutations */
        if (rngmit < opts->p_mut_geno)
          w[x][y].genome[0] = floor(rngmit*3);
        if (rngmit < opts->p_mut_geno)
          w[x][y].genome[1] = floor(rngmit*3);
        if (rngmit < opts->p_mut_geno)
          hydra->genome[0]  = floor(rngmit*3);
        if (rngmit < opts->p_mut_geno)
          hydra->genome[1]  = floor(rngmit*3);
        
        return 1;
      }
      
    }
  }
  return 0;
}


/* who will win a fight */
void _resolveFight(Opts *opts, Hydractina* other, Hydractina *self) {
  
  switch (self->morph) {
    case stolon:
      
      switch (other->morph) {
        case stolon:
          if (rngmit < opts->p_fight_ssuu)
            self->alive  = 0;
          else
            other->alive = 0;
          break;
        case fused:
          other->alive = 0;
          break;
        case unfused:
          if (rngmit < opts->p_fight_su)
            other->alive = 0;
          else
            self->alive  = 0;
          break;
      }
      break;
      
    case fused:
      if (other->morph != fused)
        self->alive = 0;
      break;
      
    case unfused:
      switch (other->morph) {
        case unfused:
          if (rngmit < opts->p_fight_ssuu)
            self->alive  = 0;
          else
            other->alive = 0;
          break;
        case fused:
          other->alive = 0;
          break;
        case stolon:
          if (rngmit < opts->p_fight_su)
            self->alive  = 0;
          else
            other->alive = 0;
          break;
      }
      break;
  }

}


/* decides who (if anyone) to fight */
int _fight(Opts *opts, World w, Hydractina* hydra, int step) {
  int i, x, y;
  int rn = floor(rngmit * 4);

  for (i = 0 ; i < 4; i ++, rn = (rn + i) % 4) {
    
    /* (0,0) is the top-left corner */
    switch (rn) {
      case 0:
        /* above */
        x = hydra->where.x;
        y = hydra->where.y - 1;
        if (y < 0)
          continue;
        break;
      case 1:
        /* right */
        x = hydra->where.x + 1;
        y = hydra->where.y;
        if (x >= (opts->size))
          continue;
        break;
      case 2:
        /* below */
        x = hydra->where.x;
        y = hydra->where.y + 1;
        if (y >= (opts->size))
          continue;
        break;
      case 3:
        /* left */
        x = hydra->where.x - 1;
        y = hydra->where.y;
        if (x < 0)
          continue;
        break;
      default:
        /* we should not get here */
        x = y = -1;
        assert(1);
    }
    
    if( w[x][y].alive &&
        w[x][y].genome[0] != hydra->genome[0] && 
        w[x][y].genome[1] != hydra->genome[1] ) {
      _resolveFight(opts, &w[x][y], hydra);
      w[x][y].round = step;
      return 1;
    }
        
  }
  return 0;
}



/* -----------------------------------------------------------------------
 *    Public Functions
 * ----------------------------------------------------------------------- */


/* create a new world for running the simulation on */
World newWorld(Opts *opts) {
  int x, y;
  
  World w = malloc(opts->size * sizeof(Hydractina *));
  for (x = 0 ; x < opts->size ; x++)
    w[x] = malloc(opts->size * sizeof(Hydractina));

  /* set all agents dead and their location */
  for (x = 0 ; x < opts->size ; x++) {
    for (y = 0 ; y < opts->size ; y++) {
      w[x][y].alive = 0;
      w[x][y].where.x = x;
      w[x][y].where.y = y;
    }
  }

  return w;
}


/* main simulation loop */
void updateWorld(Opts *opts, World w, Hash hydralist, int step) {
  int agent_n;

  for (agent_n = 0; agent_n < opts->sizesq; agent_n++)
    if (hydralist[agent_n]->alive && hydralist[agent_n]->round != step) {
      
      /* check if dying */
      if (rngmit < opts->p_death)
        hydralist[agent_n]->alive = 0;
      
      /* check if fighting */
      else if (_fight(opts, w, hydralist[agent_n], step))
        hydralist[agent_n]->round = step;
      
      /* check if fucking */
      else if (_fuck(opts, w, hydralist[agent_n]))
        hydralist[agent_n]->round = step;
      
      /* else didn't do anything this round */
      else
        hydralist[agent_n]->round = step;
    }
}


/* copies hydra into the world */
void placeHydra(World w, Hydractina *hydra) {
  int x = hydra->where.x;
  int y = hydra->where.y;
  _copyHydra(&w[x][y], hydra);
}
  

