#include <gd.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "Grafx.h"
#include "Hydractina.h"
#include "Opts.h"
#include "World.h"
#include "rngmit.h"


int main(int argv, char **argc) {
  Hydractina  hydra;
  Grafx       *g;
  Hash        hydralist;
  Opts        *opts;
  World       w;
  int         i;
  char        fname[128];
  
  rngseed(time(0));
  opts = newOpts(argv, argc);

  g = newGrafx(opts);
  w = newWorld(opts);

  if (!opts->init_state) {
	  for (i = 0; i < opts->init_num; i++) {
      hydra.alive     = 1;
      hydra.where.x   = floor( rngmit * opts->size );
      hydra.where.y   = floor( rngmit * opts->size );
      hydra.morph     = floor( rngmit * 3 );
      hydra.genome[0] = floor( rngmit * 3 );
      hydra.genome[1] = floor( rngmit * 3 );
      placeHydra(w, &hydra);
    }
  } else {
    hydra.alive     = 1;
    hydra.where.x   = floor( rngmit * opts->size );
    hydra.where.y   = floor( rngmit * opts->size );
    hydra.morph     = stolon;
    hydra.genome[0] = xx;
    hydra.genome[1] = xx;
    placeHydra(w, &hydra);

    hydra.where.x   = floor( rngmit * opts->size );
    hydra.where.y   = floor( rngmit * opts->size );
    hydra.morph     = fused;
    hydra.genome[0] = xx;
    hydra.genome[1] = yy;  
    placeHydra(w, &hydra);

    hydra.where.x   = floor( rngmit * opts->size );
    hydra.where.y   = floor( rngmit * opts->size );
    hydra.morph     = unfused;
    hydra.genome[0] = yy;
    hydra.genome[1] = yy;  
    placeHydra(w, &hydra);
  }  

  /* dump png of initial world */
  world2png(opts, w, g);
  sprintf(fname, "%s-%04d.png", opts->filename, 0);
  writepng(g, fname);
  
  for (i = 1; i <= opts->steps; i++) {
    /* sim */
    hydralist = newRandHash(opts, w);
    updateWorld(opts, w, hydralist, i);
    
    /* dump the png */
    world2png(opts, w, g);
    sprintf(fname, "%s-%04d.png", opts->filename, i);
    writepng(g, fname);
    
    /* cleanup */
    free(hydralist);
  }

  gdImageDestroy(g->im);
  return 0;
}
